/*
 * Copyright (c) 2007 Luigi Rizzo - Universita` di Pisa
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Id: ldev_stub.h,v 1.20 2008/04/08 18:56:22 luigi Exp $
 *
 * This is a header file to implement the necessary functions for a
 * device driver as a loadable FreeBSD module, and on the other
 * side provides sufficient hooks to Linux device drivers to be linked
 * and operate as expected.
 *
 * compile with the following macros
 *
 * -DDEV_NAME="video"	device name (with quotes)
 * -DDRIVER_NAME=foo	driver name (without quotes)
 * make sure ldev_devs is the array of ldev_info with vendor/prod identifiers
 * to be used for matching the device.
 */

#ifndef _LDEV_STUB_H
#define _LDEV_STUB_H

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/fcntl.h>
// #include <sys/tty.h>
#include <sys/selinfo.h>
#include <sys/poll.h>
#include <sys/uio.h>		/* uiomove */

/* probably needed later when implementing malloc */
// #include <sys/sysctl.h>
// #include <sys/malloc.h>
// #include <sys/mman.h>
// #include <vm/vm.h>
// #include <vm/pmap.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usb_port.h>

#include <linux/kernel.h>	/* DBG macro and more */
#include <linux/types.h>	/* linux types */
#include <linux/init.h>	/* linker set and module_init */
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/device.h>
#include <linux/fs.h>		/* struct file_operations */

#include <linux/usb.h>	/* this is targeted to usb... */

#define LDEV_ASLEEP      1
#define LDEV_INIT        2
#define LDEV_POLL        4

/*
 * We store here the various descriptors that linux expects,
 * and a pointer back to the FreeBSD usb device for use in the
 * emulation layer.
 * On top, there are the corresponding FreeBSD descriptors
 */
#define MAX_IF_ALTSETTINGS	16
struct linux_usb_descriptors {
        const struct file_operations *fops;
        usbd_interface_handle   iface;
        usbd_device_handle      udev;
	usbd_pipe_handle	bsd_pipe;	/* pipe used for streaming */
 
	void *mem;		/* malloced space for linux usb descriptors */
        /* linux fields */
        struct usb_interface    intf;
        struct usb_device       p_dev;  /* parent device ? */
        struct file             file;
};

/*
 * given a linux_usb_device pointer (l_dev), returns the corresponding
 * usbd_device_handle or usbd_interface_handle - assumes that l_dev is part
 * of a properly initialized struct linux_usb_descriptors
 *
 * The first macro returns the struct pointer, later to be mapped
 * to usbd_device_handler or usbd_interface_handle
 */
#define ludev_to_l_u_d(l_dev)                                   \
    ((struct linux_usb_descriptors *)container_of(l_dev,        \
                struct linux_usb_descriptors, p_dev))
#define ludev_to_udev(l_dev)    (ludev_to_l_u_d(l_dev)->udev)
#define ludev_to_iface(l_dev)   (ludev_to_l_u_d(l_dev)->iface)

/*
 * Device descriptor. Store here various information acquired through
 * during the attach phase, so that they are later available in the
 * cdevsw handlers.
 */
struct ldev_softc {
	device_t	sc_dev;		/* FreeBSD device */
	struct cdev *sc_dev_t;

	struct selinfo rsel;

	usbd_device_handle udev;	/* usb device for this interface */
	usbd_interface_handle sc_iface;	/* our interface */

	int error_status;
	int vopen;
	int state;

	struct mtx ptrlock;	/* for manipulating the buffer pointers */

	/*
	 * next is for the linux emulation structs, which must be
	 * persistent.
	 */
	struct linux_usb_descriptors	l_u_d;
	void *data_src;
	void *full_frames;	/* poll placeholder */
	/* other private data */
};

void usbd_dump_pipe(usbd_pipe_handle pipe);	/* debug */

#endif /* _LDEV_STUB_H */
